﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;

namespace Robo
{
    /// <summary>
    /// A compound comparer.
    /// </summary>
    /// <typeparam name="T">The type to compare.</typeparam>
    [ComVisible(false)]
    public sealed class CompoundComparer<T> : Comparer<T>
    {
        /// <summary>
        /// First comparer.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Security", "CA2104:DoNotDeclareReadOnlyMutableReferenceTypes")]
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public readonly IComparer<T> First;
        /// <summary>
        /// Second comparer.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Security", "CA2104:DoNotDeclareReadOnlyMutableReferenceTypes")]
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1051:DoNotDeclareVisibleInstanceFields")]
        public readonly IComparer<T> Second;
        /// <summary>
        /// Compares two T objects.
        /// </summary>
        /// <param name="x">The first object.</param>
        /// <param name="y">The second object.</param>
        /// <returns>The result from first comparer, delegating to second comparer in case of a tie.</returns>
        public override int Compare(T x, T y)
        {
            var res = First.Compare(x, y);
            if (res != 0)
                return res;
            return Second.Compare(x, y);
        }
        /// <summary>
        /// Initializes a compound comparer, using two comparers.
        /// </summary>
        /// <param name="first">The first comparer.</param>
        /// <param name="second">The second comparer.</param>
        public CompoundComparer(IComparer<T> first, IComparer<T> second)
        {
            this.First = first;
            this.Second = second;
        }
        /// <summary>
        /// Chains the comparers, recursively.
        /// </summary>
        /// <param name="comparers">The comparers.</param>
        public CompoundComparer(params IComparer<T>[] comparers)
        {
            if (comparers == null) throw new ArgumentNullException("comparers");
            if (comparers.Length < 3) throw new ArgumentException("Invalid comparer count", "comparers");
            First = comparers.First();
            if (comparers.Skip(1).Count() > 2)
                Second = new CompoundComparer<T>(comparers.Skip(1).ToArray());
            else
                Second = new CompoundComparer<T>(comparers[1], comparers[2]);
        }
    }
}
