﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Runtime.Serialization;
using System.Threading;
using System.Threading.Tasks;
using System.Linq;
using System.IO;
using System.Diagnostics;
using System.Diagnostics.Contracts;
using System.Runtime.CompilerServices;

namespace Robo
{
    /// <summary>
    /// A static class containing task-related extensions.
    /// </summary>
    public static class TaskHelper
    {
        /// <summary>
        /// Wraps a System.Threading.WaitHandle <paramref name="handle"/> in an awaitable task, using given timeout.
        /// </summary>
        /// <param name="handle">The handle to wrap.</param>
        /// <param name="timeout">An optional timeout. The default is -1 (Infinite)</param>
        /// <returns>A task signalled when the handle is set. The task returns whether or not the handle timed out.</returns>
        public static Task<bool> AsTask(this WaitHandle handle, int timeout = -1)
        {
            if (handle == null) throw new ArgumentNullException("handle");
            Contract.EndContractBlock();
            TaskCompletionSource<bool> tcs = new TaskCompletionSource<bool>();
            RegisteredWaitHandle token = null;
            var tokenReady = new ManualResetEventSlim();
            token = ThreadPool.RegisterWaitForSingleObject(
                handle,
                (state, timedOut) =>
                {
                    tokenReady.Wait();
                    tokenReady.Dispose();
                    token.Unregister(handle);
                    tcs.SetResult(!timedOut);
                },
                null,
                timeout,
                true);
            tokenReady.Set();
            return tcs.Task;
        }
        /// <summary>
        /// Gets the awaiter for the WaitHandle.
        /// </summary>
        /// <param name="handle">The handle to await.</param>
        /// <returns>The awaiter.</returns>
        public static TaskAwaiter<bool> GetAwaiter(this WaitHandle handle)
        {
            return handle.AsTask().GetAwaiter();
        }
    }

    /// <summary>
    /// Provides general extensions.
    /// </summary>
    public static class Extensions
    {
        #region Enumerator
        /// <summary>
        /// Wraps an enumerator given its enumerator functions.
        /// </summary>
        /// <typeparam name="T">The element type</typeparam>
        /// <param name="moveNext">The IEnumerator.MoveNext() function</param>
        /// <param name="getCurrent">The IEnumerator{T}.get_Current() function.</param>
        /// <param name="dispose">The (optional) IDisposable.Dispose() funcion.</param>
        /// <returns>The Enumerable.</returns>
        public static IEnumerable<T> Wrap<T>(Func<bool> moveNext, Func<T> getCurrent, Action dispose = null)
        {
            if (moveNext == null) throw new ArgumentNullException("moveNext");
            if (getCurrent == null) throw new ArgumentNullException("getCurrent");
            Contract.EndContractBlock();
            try
            {
                while (moveNext())
                    yield return getCurrent();
            }
            finally
            {
                if (dispose != null)
                    dispose();
            }
        }

        /// <summary>
        /// Wraps an T:System.Collectioons.Generic.IEnumerator{T} instance in an enumerable. Shrotcut for calling <see cref="Wrap{T}(System.Func{System.Boolean}, System.Func{T}, System.Action)"/>.
        /// </summary>
        /// <typeparam name="T">The element type.</typeparam>
        /// <param name="enumerator">The enumerator to wrap.</param>
        /// <returns>The enumerable wrapper.</returns>
        /// <seealso cref="Wrap{T}(System.Func{System.Boolean}, System.Func{T}, System.Action)"/>
        public static IEnumerable<T> Wrap<T>(this IEnumerator<T> enumerator)
        {
            if (enumerator == null) throw new ArgumentNullException("enumerator");
            Contract.EndContractBlock();
            return Wrap(enumerator.MoveNext, () => enumerator.Current, enumerator.Dispose);
        }

        /// <summary>
        /// Wraps an T:System.Collectioons.IEnumerator. Shortcut for calling <see cref="Wrap{T}(System.Func{System.Boolean}, System.Func{T}, System.Action)"/>.
        /// </summary>
        /// <param name="enumerator">The enumerator to wrap</param>
        /// <returns>The enumerable sequence</returns>
        /// <seealso cref="Wrap{T}(System.Func{System.Boolean}, System.Func{T}, System.Action)"/>
        public static IEnumerable Wrap(this IEnumerator enumerator)
        {
            if (enumerator == null) throw new ArgumentNullException("enumerator");
            Contract.EndContractBlock();
            return Wrap(enumerator.MoveNext, () => enumerator.Current, null);
        }
        #endregion
        /// <summary>
        /// Strongly typed version of <see cref="SerializationInfo.GetValue(string, System.Type)"/>
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="info"></param>
        /// <param name="name"></param>
        /// <returns></returns>
        public static T GetValue<T>(this SerializationInfo info, string name)
        {
            if (info == null) throw new ArgumentNullException("info");
            if (name == null || string.IsNullOrEmpty(name.Trim())) throw new ArgumentException("Cannot have empty name!", "name");
            Contract.EndContractBlock();
            return (T)info.GetValue(name, typeof(T));
        }
        #region Enumerable Query
        /// <summary>
        /// Delimates a collection of data.
        /// </summary>
        /// <typeparam name="TSource">The element type.</typeparam>
        /// <param name="source">The source collection</param>
        /// <param name="delimiter">The delimiter.</param>
        /// <returns>The delimited string.</returns>
        public static string Delimit<TSource>(this IEnumerable<TSource> source, string delimiter)
        {
            if (source == null) throw new ArgumentNullException("source");
            if (delimiter == null || string.IsNullOrEmpty(delimiter.Trim())) throw new ArgumentNullException("delimiter");
            Contract.EndContractBlock();
            return string.Join(delimiter, source);
        }
        /// <summary>
        /// Delimates a collection of data, using <paramref name="selector"/> to select the data.
        /// </summary>
        /// <typeparam name="TSource">The element type.</typeparam>
        /// <param name="source">The source collection.</param>
        /// <param name="selector">Selects what to delimit.</param>
        /// <param name="delimiter">The delimiter.</param>
        /// <returns>The delimited string.</returns>
        public static string Delimit<TSource>(this IEnumerable<TSource> source, Func<TSource, string> selector, string delimiter)
        {
            if (source == null) throw new ArgumentNullException("source");
            if (selector == null) throw new ArgumentNullException("selector");
            if (delimiter == null || string.IsNullOrEmpty(delimiter.Trim())) throw new ArgumentNullException("delimiter");
            return source.Select(selector).Delimit(delimiter);
        }
        /// <summary>
        /// Splits the collection, using the given delimiter.
        /// </summary>
        /// <typeparam name="TSource">The element type.</typeparam>
        /// <param name="source">The source collection.</param>
        /// <param name="delimiter">The delimiter object.</param>
        /// <returns>The delimited collection.</returns>
        /// <remarks>Delegates to <see cref="Split{TSource}(IEnumerable{TSource}, TSource, IEqualityComparer{TSource})"/></remarks>
        /// <exception cref="ArgumentNullException">Source is null.</exception>
        public static IEnumerable<IEnumerable<TSource>> Split<TSource>(this IEnumerable<TSource> source, TSource delimiter)
        {
            if (source == null) throw new ArgumentNullException("source");
            Contract.EndContractBlock();
            return source.Split(delimiter, EqualityComparer<TSource>.Default);
        }
        /// <summary>
        /// Splits the collection, using the given delimiter and equality comparer.
        /// </summary>
        /// <typeparam name="TSource">The element type.</typeparam>
        /// <param name="source">The source collection.</param>
        /// <param name="delimiter">The delimiter object.</param>
        /// <param name="comparer">The equality comparer.</param>
        /// <returns>The delimited collection.</returns>
        public static IEnumerable<IEnumerable<TSource>> Split<TSource>(this IEnumerable<TSource> source, TSource delimiter, IEqualityComparer<TSource> comparer)
        {
            comparer = comparer ?? EqualityComparer<TSource>.Default;
            using (IEnumerator<TSource> rator = source.GetEnumerator())
            {
                bool flag = true;
                IEnumerable<TSource> ret = null;
                do
                {
                    if (flag)
                    {
                        flag = false; continue;
                    }
                    yield return ret;
                }
                while (SplitBody(rator, o => comparer.Equals(o, delimiter), out ret));
            }
        }
        [DebuggerHidden]
        [DebuggerStepThrough]
        private static bool SplitBody<TSource>(IEnumerator<TSource> rator,Predicate<TSource> pred, out IEnumerable<TSource> ret)
        {
            Contract.Requires(rator != null);
            List<TSource> list = new List<TSource>();
            bool flag = false;
            while (rator.MoveNext())
            {
                var item = rator.Current;
                list.Add(item);
                if (pred(item))
                {
                    flag = true;
                    break;
                }
            }
            ret = list.Select(x => x);
            return flag;
        }
        #endregion
    }
}
